﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using MFORMATSLib;

namespace SampleZoom
{
    public partial class Form1 : Form
    {
        // Create a MF classes
        private MFReaderClass m_objReader = new MFReaderClass();
        private MFPreviewClass m_objPreview = new MFPreviewClass();
        private MFRendererClass m_objRenderer = new MFRendererClass();
        private MFSinkClass m_objMFSink = new MFSinkClass();
        private MFFactoryClass factory = new MFFactoryClass();
        private Thread m_threadWorker;
        private Object m_objLock = new Object();
        private bool m_bWork;
        private bool m_bPlaybackTriggerOn = true;
        private int m_nMousePosX;
        private int m_nMousePosY;
        private const string PAUSE_PLAYBACK = "pause=true";
        private bool m_bDraw;    // mouse click to preview
        private bool m_bMouseMoving;   // mouse moving flag

        // For frame capture
        private string m_strCaptureFramePath;

        private M_AV_PROPS m_avProps;

        public Form1()
        {
            InitializeComponent();
        }
        

        private MF_RECT m_rcOverlay;
        private string m_strTextContent;
        private double m_dblTextSize;

        private string m_strTextConfig;
        private CancellationTokenSource m_tknCancelSource;
        private void Form1_Load(object sender, EventArgs e)
        {
            FillVideoFormats();

            // Fill renderer devices
            FillRenderers();

            // Enable video preview
            m_objPreview.PreviewWindowSet("", panelPreview.Handle.ToInt32());
            m_objPreview.PreviewEnable("", 0, 1);

            m_objPreview.PropsSet("maintain_ar", "false");

            // Create thread worker
            System.Windows.Forms.Control.CheckForIllegalCrossThreadCalls = false;   //Need for use Combobox in multithreading 
           
            m_tknCancelSource = new CancellationTokenSource();
            m_threadWorker = new Thread(() => thread_DoWork(m_tknCancelSource.Token));
            m_threadWorker.Name = "thread_DoWork";
            m_threadWorker.Start();
        }

        private void thread_DoWork(CancellationToken token)
        {
            MFFrame mask;
            factory.MFFrameCreateFromFile(@"dark_circle2.png", out mask, "");

            factory.MFFrameCreateFromFile(@"dark_circle.png", out MFFrame newmask, "");
            while (!token.IsCancellationRequested)
            {
                if (m_bWork)
                {
                    string strCapturePath = null;
                    string strFrameGetProps = String.Empty;
                    string strTextContent = String.Empty;
                    string strTextConfig = String.Empty;
                    double dblTextSizelocked = 0.0;
                    MF_RECT rcOverlay = new MF_RECT();
                    lock (m_objLock)
                    {
                        rcOverlay = m_rcOverlay;
                        strTextContent = m_strTextContent;
                        strTextConfig = m_strTextConfig;
                        dblTextSizelocked = m_dblTextSize;
                        strCapturePath = m_strCaptureFramePath;
                        m_strCaptureFramePath = String.Empty;

                        if (m_bPlaybackTriggerOn == false)
                        {
                            strFrameGetProps = PAUSE_PLAYBACK;
                        }
                        else if (isReverse)
                        {
                            strFrameGetProps = " reverse=true";
                        }

                        if (isFrameByFrame)
                        {
                            m_bPlaybackTriggerOn = false;
                            PlayPauseButton.Text = m_bPlaybackTriggerOn ? BUTTON_PAUSE : BUTTON_PLAY;
                        }

                        isReverse = false;
                        isFrameByFrame = false;
                    }
                    MFFrame mFrame;
                    try
                    {
                        m_objReader.SourceFrameConvertedGetByTime(ref m_avProps, -1, -1, out mFrame, strFrameGetProps); //playbackTriggerOn? String.Empty: PAUSE_PLAYBACK
                    }
                    catch (Exception)
                    {
                        return;
                    }

                    if (!String.IsNullOrEmpty(strTextContent))
                    {
                        // Clone frame before overlay 
                        MFFrame mFrameClone = null;
                        try
                        {
                            mFrame.MFClone(out mFrameClone, eMFrameClone.eMFC_Full, eMFCC.eMFCC_Default);

                            mFrameClone.MFPrint(strTextContent, dblTextSizelocked, ref rcOverlay, eMFTextFlags.eMFT_WordBreaks, strTextConfig);
                            Marshal.ReleaseComObject(mFrame);
                            mFrame = mFrameClone;
                        }
                        catch (System.Exception)
                        {
                            if (mFrameClone != null)
                                Marshal.ReleaseComObject(mFrameClone);
                        }
                    }
                    if (m_bDraw)
                    {
                        //MFFrame mFrameDraw = m_thePainter.DrawFrame(m_bMouseMoving, mFrame, panelPreview.Width, panelPreview.Height, m_nMousePosX, m_nMousePosY);
                        mFrame.MFClone(out MFFrame myFrameClone, eMFrameClone.eMFC_Full, eMFCC.eMFCC_Default);
                        myFrameClone.MFAVPropsGet(out M_AV_PROPS myProps, out int samples);
                        MFFrame cutFrame;
                        int x = Math.Abs(myProps.vidProps.nWidth * m_nMousePosX / panelPreview.Width);
                        int y = Math.Abs(myProps.vidProps.nHeight * m_nMousePosY / panelPreview.Height);
                        var diffX = x + 100 > myProps.vidProps.nWidth ? x + 100 - myProps.vidProps.nWidth : (x - 100 < 0 ? x-100 : 0);
                        var diffY = y + 100 > Math.Abs(myProps.vidProps.nHeight) ? y + 100 - Math.Abs(myProps.vidProps.nHeight) : (y - 100 < 0 ? y-100 : 0);

                        // rect 200x200
                        tagRECT rect = new tagRECT()
                        {
                            top = y - 100 - diffY,
                            bottom = y + 100 > Math.Abs(myProps.vidProps.nHeight) ? Math.Abs(myProps.vidProps.nHeight) : y + 100,
                            left = x - 100  - diffX,
                            right = x + 100 > myProps.vidProps.nWidth ? myProps.vidProps.nWidth : x + 100,
                        };
                        myFrameClone.MFCut(0, ref rect, out cutFrame);
                        double zoomFactor = 500 / 200; // the target rectangle width against the original cut size
                        MF_RECT newRect = new MF_RECT()
                        {
                            dblHeight = 500,
                            dblWidth = 500,
                            dblPosX = rect.left - 500 / (zoomFactor * 2), // + 100 > Math.Abs(myProps.vidProps.nHeight) ? Math.Abs(myProps.vidProps.nHeight) - 100 : rect.top + 100,
                            dblPosY = rect.top - 500 / (zoomFactor * 2)
                        };
                        myFrameClone.MFOverlayRect(cutFrame, mask, ref newRect, eMFOverlayFlags.eMFOF_Crop, 1, "scale_type=2", "");
                        myFrameClone.MFOverlayRect(newmask, null, ref newRect, eMFOverlayFlags.eMFOF_Crop, 1, "scale_type=2", "");
                        Marshal.ReleaseComObject(cutFrame);
                        //Marshal.ReleaseComObject(convertedFrame);
                        Marshal.ReleaseComObject(mFrame);
                        mFrame = myFrameClone;
                    }

                    if (!String.IsNullOrEmpty(strCapturePath))
                    {
                        mFrame.MFVideoSaveToFile(strCapturePath);
                    }

                    //Send to Sink
                    SendToSink(mFrame);

                    // Send frame to Renderer
                    bool bUseRenderer = SendToRenderer(mFrame);

                    // Send frame to preview (if frame renderer to device, do not wait in preview)
                    try
                    {
                        m_objPreview.ReceiverFramePut(mFrame, bUseRenderer ? 0 : -1, "");
                    }
                    catch { }

                    // Release objects
                    Marshal.ReleaseComObject(mFrame);
                }
                else
                {
                    // For prevent CPU overload if something wrong happened 
                    Thread.Sleep(1);
                }
            }
        }

        private void SendToSink(MFFrame pFrame)
        {
            if (checkBoxVirtDev.Checked)
            {
                if (m_objMFSink != null)
                {
                    lock (m_objLock)
                    {
                        // Send pFrame to Sink
                        try
                        {
                            m_objMFSink.ReceiverFramePut(pFrame, -1, "");
                        }
                        catch
                        { }
                    }
                }
            }
        }

        private bool SendToRenderer(MFFrame pFrame)
        {
            if (checkBoxRenderer.Checked && comboBoxRenderers.SelectedItem != null && m_objRenderer != null)
            {
                try
                {
                    // Send pFrame to renderer
                    m_objRenderer.ReceiverFramePut(pFrame, -1, "");
                    return true;
                }
                catch { }
            }

            return false;
        }

        private void FillVideoFormats()
        {
            comboBoxVideoFormat.Items.Clear();

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_Custom);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_NTSC);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_NTSC_2398);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_NTSC_16x9);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_PAL);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_PAL_16x9);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD720_50p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD720_5994p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD720_60p);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_2398p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_24p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_25p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_2997p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_30p);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_50i);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_5994i);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_60i);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_50p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_5994p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_HD1080_60p);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_DCI_2398p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_DCI_24p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_DCI_25p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_DCI_50p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_DCI_5994p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_DCI_60p);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_2398p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_24p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_2K_25p);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_DCI_2398p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_DCI_24p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_DCI_25p);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_UHD_50i);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_UHD_5994i);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_UHD_60i);

            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_UHD_50p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_UHD_5994p);
            comboBoxVideoFormat.Items.Add(eMVideoFormat.eMVF_4K_UHD_60p);

            comboBoxVideoFormat.SelectedIndex = 0;

            m_avProps = new M_AV_PROPS();
            m_avProps.vidProps.eVideoFormat = (eMVideoFormat)comboBoxVideoFormat.SelectedItem;
        }

        private void FillRenderers()
        {
            comboBoxRenderers.Items.Clear();

            int nCount = 0;

            m_objRenderer.DeviceGetCount(eMFDeviceType.eMFDT_Video, out nCount);

            if (nCount > 0)
            {
                comboBoxRenderers.Enabled = true;
                checkBoxRenderer.Enabled = true;

                for (int y = 0; y < nCount; y++)
                {
                    string strName;
                    int _pbBusy;
                    m_objRenderer.DeviceGetByIndex(eMFDeviceType.eMFDT_Video, y, out strName, out _pbBusy);
                    comboBoxRenderers.Items.Add(strName);
                }

                // Choose 0 index of renderer
                comboBoxRenderers.SelectedIndex = 0;

            }
        }

        private void buttonOpenFile_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                textBoxPath.Text = ofd.FileName;
                TryToOpenFile(ofd.FileName);
            }
        }

        private void panel1_MouseDown(object sender, MouseEventArgs e)
        {
            m_bDraw = true;
            
        }

        private void panel1_MouseUp(object sender, MouseEventArgs e)
        {
            m_bDraw = false;
        }

        private void panel1_MouseMove(object sender, MouseEventArgs e)
        {
            if (m_bDraw)
            {
                m_nMousePosX = e.X;
                m_nMousePosY = e.Y;
                m_bMouseMoving = true;
            }
            else
            {
                m_bMouseMoving = false;
            }
        }

        private void renderersCombobox_SelectedIndexChanged(object sender, EventArgs e)
        {
            lock (m_objLock)
            {
                if (checkBoxRenderer.Checked)
                {
                    string name;
                    int nBusy;
                    m_objRenderer.DeviceGetByIndex(eMFDeviceType.eMFDT_Video, comboBoxRenderers.SelectedIndex, out name,
                        out nBusy);

                    if (nBusy < 1)
                    {
                        m_objRenderer.DeviceClose();
                        m_objRenderer.DeviceSet(eMFDeviceType.eMFDT_Video, comboBoxRenderers.SelectedIndex, "");
                    }
                }
            }
        }

        private void panel1_Resize(object sender, EventArgs e)
        {
            m_objPreview.PreviewWindowSet("", panelPreview.Handle.ToInt32());
        }

        private void checkBox_renderer_CheckedChanged(object sender, EventArgs e)
        {
            if (!checkBoxRenderer.Checked)
                m_objRenderer.DeviceClose();
            else
            {
                string name;
                int nBusy;
                m_objRenderer.DeviceGetByIndex(eMFDeviceType.eMFDT_Video, comboBoxRenderers.SelectedIndex, out name,
                    out nBusy);

                if (nBusy < 1)
                {
                    m_objRenderer.DeviceSet(eMFDeviceType.eMFDT_Video, comboBoxRenderers.SelectedIndex, "");

                }
                else
                {
                    checkBoxRenderer.Checked = false;
                    MessageBox.Show("Device is busy.");
                }
            }
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            m_tknCancelSource.Cancel();

            m_threadWorker.Join(1000);

            m_objReader.ReaderClose();

            m_objMFSink.SinkClose();
        }
        

        private const string BUTTON_PLAY = "Continue Playback";
        private const string BUTTON_PAUSE = "Pause Playback";
        private const string ERROR_OPEN_FILE = "Error open file:";
        private void PlayPause(object sender, EventArgs e)
        {
            isFrameByFrame = false;
            isReverse = false;
            m_bPlaybackTriggerOn = !m_bPlaybackTriggerOn;
            PlayPauseButton.Text = m_bPlaybackTriggerOn ? BUTTON_PAUSE : BUTTON_PLAY;
        }

        private void textBoxPath_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                TryToOpenFile(textBoxPath.Text);
            }
        }

        private void TryToOpenFile(string pathToFile)
        {
            try
            {
                m_objReader.ReaderOpen(pathToFile, "loop=true");
            }

            catch (Exception ex)
            {
                MessageBox.Show(ERROR_OPEN_FILE + pathToFile + Environment.NewLine + ex.Message);
                return;
            }
            m_bWork = true;
            PlayPauseButton.Enabled = true;
        }

        private bool isReverse;
        private bool isFrameByFrame;
        private void FrameFWD_Click(object sender, EventArgs e)
        {
            m_bPlaybackTriggerOn = true;
            isFrameByFrame = true;
            isReverse = false;
        }

        private void FrameBack_Click(object sender, EventArgs e)
        {
            m_bPlaybackTriggerOn = true;
            isFrameByFrame = true;
            isReverse = true;
        }

        private void comboBoxVideoFormat_SelectedIndexChanged(object sender, EventArgs e)
        {
            m_avProps.vidProps.eVideoFormat = (eMVideoFormat)comboBoxVideoFormat.SelectedItem;
        }

        private void comboBox_DropDown(object sender, EventArgs e)
        {
            ComboBox senderComboBox = (ComboBox)sender;
            int newWidth = senderComboBox.DropDownWidth;
            int width = senderComboBox.DropDownWidth;
            System.Drawing.Font font = senderComboBox.Font;
            System.Drawing.Graphics g = senderComboBox.CreateGraphics();
            int vertScrollBarWidth = (senderComboBox.Items.Count > senderComboBox.MaxDropDownItems) ?
              System.Windows.Forms.SystemInformation.VerticalScrollBarWidth : 0;
            foreach (object item in ((ComboBox)senderComboBox).Items)
            {
                string s = senderComboBox.GetItemText(item);
                newWidth = (int)g.MeasureString(s, font).Width + vertScrollBarWidth;
                if (width < newWidth)
                {
                    width = newWidth;
                }
            }
            senderComboBox.DropDownWidth = width;
        }

        private void checkBoxVirtDev_CheckedChanged(object sender, EventArgs e)
        {
            lock (m_objLock)
            {
                if (checkBoxVirtDev.Checked)
                {
                    m_objMFSink.SinkInit("MFReaderSink", "MFormats VSource", null);
                }
                else
                {
                    m_objMFSink.SinkClose();
                }
            }
        }

        private void buttonCaptureFrame_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveDialog = new SaveFileDialog();
            saveDialog.Filter = "PNG Files(*.png)|*.png|JPEG Files(*.jpg)|*.jpg";
            saveDialog.DefaultExt = "png";
            if (saveDialog.ShowDialog() == DialogResult.OK)
            {
                lock (m_objLock)
                {
                    m_strCaptureFramePath = saveDialog.FileName;
                }
            }

        }
    }
}
